clc; clear all; close all

% Load Data
load cart3_reg3_10states_14mps

% Indices for states, inputs, and outputs:
%   q = [Tower 1st F/A; Rotor Pos; Blade1 1st Flap; Blade2; Blade3]
%   x = [q; qdot]
%   u = [Blade1 Pitch; Blade2; Blade3; Gen Torque];
%   y = [Rotor Speed; Blade1 Root Moment; Blade2; Blade3];
% This ordering corresponds to the paper.
% Note that the data has the disturbances before the control inputs.
ndof = 5;
nx = 2*ndof;
nu = 4;
idxu = [3 4 5 2];
ny = 4;
idxy = 1:4;

% Grab state matrices for PLTV system
A = AMat;
B = BMat(:,idxu,:);
C = CMat(idxy,:,:);
D = DMat(idxy,idxu,:);

% MBC tranxformation matrices
psi = 0:2:358;
Np = length(psi);

M = zeros(3,3,Np);
Minv = M;

Mu = zeros(nu,nu,Np);
Muinv = Mu;

My = zeros(ny,ny,Np);
Myinv = My;

Mx = zeros(nx,nx,Np);
Mxinv = Mx;
Mxdot = zeros(nx,nx,Np);
for i=1:Np
    % Rotor Angle
    p = psi(i)*pi/180;
    
    % Single MBC matrix
    Mi = [1 sin(p) cos(p); ...
        1 sin(p+2*pi/3) cos(p+2*pi/3); ...
        1 sin(p+4*pi/3) cos(p+4*pi/3)];
    
    % Flip columns to correspond to MBC3 code
    Mi = Mi(:,[1 3 2]);
    
    % Store M and its inverse
    M(:,:,i) = Mi;
    Minv(:,:,i) = inv( Mi );

    % Derivative of Mi
    Midot = Omega(i)*[0 cos(p) -sin(p); ...
        0 cos(p+2*pi/3) -sin(p+2*pi/3); ...
        0 cos(p+4*pi/3) -sin(p+4*pi/3)];
    Midot = Midot(:,[1 3 2]);
    
    % Second Derivative of Mi (Assuming Omegadd = 0)
    Midd = Omega(i)^2*[0 -sin(p) -cos(p); ...
        0 -sin(p+2*pi/3) -cos(p+2*pi/3); ...
        0 -sin(p+4*pi/3) -cos(p+4*pi/3)];
    Midd = Midd(:,[1 3 2]);
    
    % I/O Tranxformationx
    Mu(:,:,i) = blkdiag(Mi,1);
    Muinv(:,:,i) = inv( Mu(:,:,i) );
    My(:,:,i) = blkdiag(1,Mi);
    Myinv(:,:,i) = inv( My(:,:,i) );
    
    % State Tranxformation
    Mq = blkdiag(eye(2),Mi);
    Mqdot = blkdiag(zeros(2),Midot);
    Mqdd = blkdiag(zeros(2),Midd);
    Mx(:,:,i) = [ Mq zeros(ndof); Mqdot Mq];
    Mxinv(:,:,i) = inv(Mx(:,:,i));
    Mxdot(:,:,i) = [ Mqdot zeros(ndof); Mqdd Mqdot];    
end


%% Optimization Initial Point: MBC+Averaged Matrices
% Method 2: MBC + Average
A0tmp = zeros(nx,nx,Np);
B0tmp = zeros(nx,nu,Np);
C0tmp = zeros(ny,nx,Np);
D0tmp = zeros(ny,nu,Np);
for i=1:Np
    A0tmp(:,:,i) = Mxinv(:,:,i)*( A(:,:,i)*Mx(:,:,i) - Mxdot(:,:,i) );
    B0tmp(:,:,i) = Mxinv(:,:,i)*B(:,:,i)*Mu(:,:,i);
    C0tmp(:,:,i) = Myinv(:,:,i)*C(:,:,i)*Mx(:,:,i);    
    D0tmp(:,:,i) = Myinv(:,:,i)*D(:,:,i)*Mu(:,:,i);
end
A0 = mean(A0tmp,3);
B0 = mean(B0tmp,3);
C0 = mean(C0tmp,3);
D0 = mean(D0tmp,3);
x0=[A0 B0; C0 D0];
% the first nx/2 lines are constant 0s and 1s
x0=x0(nx/2+1:end,:);
x0=x0(:);

% MBC+MinPeakVariation
Amax=-Inf*ones(nx,nx);
Amin=Inf*ones(nx,nx);
Bmax=-Inf*ones(nx,nu);
Bmin=Inf*ones(nx,nu);
Cmax=-Inf*ones(ny,nx);
Cmin=Inf*ones(ny,nx);
Dmax=-Inf*ones(ny,nu);
Dmin=Inf*ones(ny,nu);
for k=1:Np
    tmp=squeeze(A0tmp(:,:,k));
    idx=tmp>Amax;
    Amax(idx)=tmp(idx);
    idx=tmp<Amin;
    Amin(idx)=tmp(idx);
    
    tmp=squeeze(B0tmp(:,:,k));
    idx=tmp>Bmax;
    Bmax(idx)=tmp(idx);
    idx=tmp<Bmin;
    Bmin(idx)=tmp(idx);
    
    tmp=squeeze(C0tmp(:,:,k));
    idx=tmp>Cmax;
    Cmax(idx)=tmp(idx);
    idx=tmp<Cmin;
    Cmin(idx)=tmp(idx);  
    
    tmp=squeeze(D0tmp(:,:,k));
    idx=tmp>Dmax;
    Dmax(idx)=tmp(idx);
    idx=tmp<Dmin;
    Dmin(idx)=tmp(idx);  
end
A0inf=(Amax+Amin)/2;
B0inf=(Bmax+Bmin)/2;
C0inf=(Cmax+Cmin)/2;
D0inf=(Dmax+Dmin)/2;

clear A0tmp B0tmp C0tmp D0tmp;
clear Amax Amin Bmax Bmin Cmax Cmin Dmax Dmin;

%% Optimization Code
% options
info.coststr='inf'; % 'inf': inf-norm minimization over matrix elements
                          % 'frobenius'
info.inputScaling=1;
info.outputScaling=1;

% System Matrices
sm=[A B; C D];
% info about the system
info.nx=size(A,1);
info.nu=size(B,2);
info.ny=size(C,1);
info.psi=psi;
info.npsi=Np;                   
% MBC matrices
mbc.Mu=Mu;
mbc.Muinv=Muinv;
mbc.My=My;
mbc.Myinv=Myinv;
mbc.Mx=Mx;
mbx.Mxinv=Mxinv;
mbc.Mxdot=Mxdot;
mbc.Ms1=[Mx zeros(nx,nu,180); zeros(ny,nx,180) My];
mbc.Ms2=[Mxinv zeros(nx,nu,180); zeros(ny,nx,180) Muinv];
tmp=zeros([size(A)]);
for k=1:Np
    tmp(:,:,k)=Mxdot(:,:,k)*Mxinv(:,:,k);
end
mbc.Ms3=[tmp zeros(nx,nu,180); zeros(ny,nx+nu,180)];
clear tmp;

options = optimset('Display','iter-detailed',...
                    'Diagnostics', 'on',...
                    'Algorithm','active-set',...
                    'MaxFunEvals',numel(x0)*500,...
                    'PlotFcns',{@optimplotx,
                                @optimplotfunccount,
                                @optimplotfval,
                                @optimplotconstrviolation,
                                @optimplotstepsize,
                                @optimplotfirstorderopt});
%                     'AlwaysHonorConstraints','bounds',...
%                     'DiffMaxChange',0.05,...
%                     'DiffMinChange',0.001,...
%                     'TypicalX',ones(size(x0)),...
%                     'MaxFunEvals',900,...
%                     'OutputFcn',outputfunh,...

if info.inputScaling % scale inputs, if desired
   info.inputScale=x0;
   x0=ones(size(x0));
end    
if info.outputScaling % implement this for constraints as well?
   info.outputScale=1;
   info.outputScale=costFun(x0,sm,mbc,info); 
end
    
costf=@(x)costFun(x,sm,mbc,info);
constf=@(x)constFun(x,sm,mbc,info);
tic
[x,fval,exitflag] = fmincon(costf,x0,[],[],[],[],[],[],constf,options);
toc
if info.inputScaling % un-do scaling
   x=x.*info.inputScale;
end  

%% Results
F=[zeros(info.nx/2) eye(info.nx/2) zeros(info.nx/2,info.nu);
   reshape(x,info.nx/2+info.ny,info.nx+info.nu)]; % F=[A B; C D]
A1=F(1:info.nx,1:info.nx);
B1=F(1:info.nx,info.nx+1:end);
C1=F(info.nx+1:end,1:info.nx);
D1=F(info.nx+1:end,info.nx+1:end);

sys0=ss(A0,B0,C0,D0); %% MBC+Averaging
sys1=ss(A1,B1,C1,D1); %% fmincon

figure(2);
sigma(sys0,'b',sys1,'r--',{10^-2 10^2});
legend('MBC+Averaging','fmincon','location','best')

tstep=0:0.01:10;
[yy0,tt0,xx0] = step(pi/180*sys0(:,1),tstep);
[yy1,tt1,xx1] = step(pi/180*sys1(:,1),tstep);
figure(3);
plot(tt0,yy0(:,1),'b',...
     tt1,yy1(:,1),'r--');
xlabel('Time [s]');
ylabel('Delta RotSp [rpm]');
title('Response To 1deg Step Input on Collective Pitch Input')
legend('MBC+Averaging','fmincon','location','best')
     
garyfyFigure([2 3]);

% test inf-norm minimization through approach 2 and 1
if strcmp(info.coststr,'inf')
   % if >0 then Approach 1 and 2 are not equal for inf-norm minimization
   norm(A1-A0inf) 
end    